from shapely.geometry import Polygon
import geopandas as gpd
from h3 import h3
import folium
import logging


logging.basicConfig(
    format=' % (asctime)s | %(name)s | %(levelname)s | %(message)s', level=40)


""" when useing folium you need to reproject to 4326
import geopandas as gpd
df = gpd.read_file(data)

"""


def visualize_hexagons(hexagons, color="red", folium_map=None):
    """
    hexagons is a list of hexcluster. Each hexcluster is a list of hexagons. 
    eg. [[hex1, hex2], [hex3, hex4]]

    source: https://nbviewer.jupyter.org/github/uber/h3-py-notebooks/blob/master/notebooks/usage.ipynb
    """
    polylines = []
    lat = []
    lng = []
    for _hex in hexagons:
        polygons = h3.h3_set_to_multi_polygon([_hex], geo_json=False)
        # flatten polygons into loops.
        outlines = [loop for polygon in polygons for loop in polygon]
        polyline = [outline + [outline[0]] for outline in outlines][0]
        lat.extend(map(lambda v: v[0], polyline))
        lng.extend(map(lambda v: v[1], polyline))
        polylines.append(polyline)

    if folium_map is None:
        m = folium.Map(location=[sum(lat)/len(lat), sum(lng) /
                                 len(lng)], zoom_start=13, tiles='cartodbpositron')
    else:
        m = folium_map
    for polyline in polylines:
        my_PolyLine = folium.PolyLine(
            locations=polyline, weight=8, color=color)
        m.add_child(my_PolyLine)
    return m



def visualize_polygon(polyline, color):
    """
    source: https://nbviewer.jupyter.org/github/uber/h3-py-notebooks/blob/master/notebooks/usage.ipynb
    """
    polyline.append(polyline[0])
    lat = [p[0] for p in polyline]
    lng = [p[1] for p in polyline]
    m = folium.Map(location=[sum(lat)/len(lat), sum(lng) /
                             len(lng)], zoom_start=13, tiles='cartodbpositron', crs='EPSG3006')   # default 'EPSG3857'
    my_PolyLine = folium.PolyLine(locations=polyline, weight=8, color=color)
    m.add_child(my_PolyLine)
    return m


def build_hexs(gdf: gpd.GeoDataFrame, APERTURE_SIZE: int = 8):
    """
    source: https://geographicdata.science/book/data/h3_grid/build_sd_h3_grid.html
    """
    hexs = h3.polyfill(
        gdf.geometry[0].__geo_interface__, APERTURE_SIZE, geo_json_conformant=True)
    return hexs


def gdf_to_h3_geojson(gdf: gpd.GeoDataFrame):

    assert gdf.crs is not None
    assert gdf.crs != ""

    logging.info(f"gdf.crs: {gdf.crs}")
    df = gdf.to_crs(epsg='4326')
    #  this is a string
    js = df.to_json()

    return js


def polygonise(hex_id):
    """
    source: https://geographicdata.science/book/data/h3_grid/build_sd_h3_grid.html
    """
    return Polygon(h3.h3_to_geo_boundary(
        hex_id, geo_json=True))


def polygonize_hexagons(gdf: gpd.GeoDataFrame, APERTURE_SIZE: int = 8, crs: str = "EPSG:4326"):
    """

    https://geographicdata.science/book/data/h3_grid/build_sd_h3_grid.html
    """

    GeoJSON_polygon = gdf_to_h3_geojson(gdf=gdf)

    hexs = h3.polyfill(GeoJSON_polygon, APERTURE_SIZE,
                       geo_json_conformant=True)

    all_polys = gpd.GeoSeries(list(map(polygonise, hexs)),
                              index=hexs,
                              crs=crs
                              )
    return all_polys


def show_map(
    geojson_result,
    center_location: list,
    zoom_start: float = 5.5,
    tiles: str = "cartodbpositron",
    attr: str = '''© <a href="http://www.openstreetmap.org/copyright">
                          OpenStreetMap</a>contributors ©
                          <a href="http://cartodb.com/attributions#basemaps">
                          CartoDB</a>''',
    name: str = 'Map'
                    
):
    """[summary]

    Args:
        center_location (List): [description]
        zoom_start (float, optional): [description]. Defaults to 5.5.
        tiles (str, optional): [description]. Defaults to "cartodbpositron".
        attr (str, optional): [description]. Defaults to '''© <a href="http://www.openstreetmap.org/copyright"> OpenStreetMap</a>contributors © <a href="http://cartodb.com/attributions#basemaps"> CartoDB</a>'''.

    Returns:
        [type]: [description]
    """

    logging.info(geojson_result)

    fmap = folium.Map(location=center_location,
                      zoom_start=zoom_start,
                      tiles=tiles,
                      attr=attr
                      )

    folium.GeoJson(
        geojson_result,
        style_function=lambda feature: {
            'fillColor': None,
            'color': ("green"),
            'weight': 1,
            'fillOpacity': 0.05
        },
        name=name,
    ).add_to(fmap)

    return fmap



import numpy as np
import geopandas as gpd
from shapely.geometry import Polygon



def hexagon_corners(center, w):
    """
    w -> distance_between polygons (width in the pointy arrangement)
    https://github.com/McJazzy/hexagonpy/blob/master/image.py
    """
    #x = center[0]
    #y = center[1]
    
    # he inner radius is equal to (sqrt(3) /2) times the outer radius
    #inner_radius = (math.sqrt(3) / 2) * size
    # w = 2*inner_radius
    # w = distance_between_hexagons
    # size = (distance_between_hexagons / 2) / math.sqrt(3)
     
    size = w / np.sqrt(3)
    h = 2 * size
    
    angles_in_radians = np.deg2rad([60 * i + 30 for i in range(6)])
    x = center[0] + size * np.cos(angles_in_radians)
    y = center[1] + size * np.sin(angles_in_radians)
    
    ## points = np.round(np.vstack([x, y]).T)
    corners_points = np.vstack([x, y]).T
    return corners_points
    

    #return [ 
    #    (x - w / 2, y - h / 4),
    #    (x, y - h /2),
    #    (x + w/2, y - h / 4),
    #    (x + w/2, y + h/4),
    #    (x, y + h/2),
    #    (x - w / 2, y + h/4)
    #]

    

def rectangle_corners(center, w, h):
    x = center[0]
    y = center[1]

    return [
        (x - w/2, y - h/2),
        (x + w/2 , y - h/2),
        (x +w/2, y + h/2),
        (x -w /2, y + h/2)
    ]



def build_hexagon_grid(roi:gpd.GeoDataFrame, seed_point:tuple, distance_between_hexagons: int = 500)->gpd.GeoDataFrame:
    """Generates an hexagon grid, starting from the seed_point (x,y) 
    
    :params roi:
    :params seed_point:
    :params distance_between_hexagons: Equals to the width of a pointy hexagon.
    
    """
    
    w = distance_between_hexagons
    # size = (distance_between_hexagons / 2) / math.sqrt(3)
    hexagon_size = w / np.sqrt(3)
    h = 2 * hexagon_size
    
    roi_width = int(roi.bounds['maxx'][0]) - int(roi.bounds['minx'][0])
    roi_height = int(roi.bounds['maxy'][0]) - int(roi.bounds['miny'][0])
    # 'width: {}, height: {}'.format(roi_width, roi_height)
    
    # number of hexagons horizontally and vertically
    # The size is given as a 2-tuple (width, height)
    num_hor = int(roi_width / w) + 2
    num_ver = int(roi_height / h * 4 / 3) + 2
    num_hor, num_ver

    #
    hexes = []
    for j in range(-num_ver, num_ver):
        for i in range(0, num_hor): # num_hor*num_ver
            column = i % num_hor
            row = j # i // num_hor
            even = row % 2  # the even rows of hexagons has w/2 offset on the x-axis compared to odd rows.    

            p = Polygon(hexagon_corners(( (column*w + even * w/2)+ seed_point[0], (row*h * 3/4)+seed_point[1]), w = w))
            if p.centroid.within(roi.unary_union):    # roi.envelope.unary_union    
                hexes.append(p)

    # hexagon_grid            
    return gpd.GeoDataFrame(index=range(0, len(hexes)), crs=roi.crs, geometry=hexes)


def generate_random_seed_point_within_polygon(roi:gpd.GeoDataFrame)->tuple:
    """
    Generate one random.uniform seed, within the roi polygon.
    Used to feed the `seed_point` to the hexagon grid.
    
    return
    """
    x_min, y_min, x_max, y_max = roi.total_bounds
    # set sample size
    n = 1
    gdf_points = None
    is_in_polygon = False
    while not is_in_polygon:
        # We iterate as the random ppoint may fall outside the target polygon but within its boundaries.
        # generate random data within the bounds, assumes Euclidean distances.
        x = np.random.uniform(x_min, x_max, n)
        y = np.random.uniform(y_min, y_max, n)

        # convert them to a points GeoSeries
        gdf_points = gpd.GeoSeries(gpd.points_from_xy(x, y))
        # only keep those points within polygons
        gdf_points = gdf_points[gdf_points.within(roi.unary_union)]
        if len(gdf_points)>0:
            is_in_polygon = True
    # '{}, {}'.format(int(gdf_points[gdf_points.keys()[0]].x), int(gdf_points[gdf_points.keys()[0]].y))
    return (int(gdf_points[gdf_points.keys()[0]].x), int(gdf_points[gdf_points.keys()[0]].y))


from osgeo import gdal
import rasterio
import geopandas as gpd
from shapely.geometry import Point
from typing import Union, Callable, List, Tuple
import logging
import os


logging.basicConfig(
    format=' % (asctime)s | %(name)s | %(levelname)s | %(message)s', level=40)



def geometry_centroids(
    gdf:gpd.GeoDataFrame,
    round_up_crs_list:list = ['epsg:3006', 'epsg:3035'] )-> List[Tuple[Point]]:
    """Returns the `centroids` of the input `GeoDataFrame`.
    WARNING: Do not add `EPSG:4326` (WGS84) or non projected coordinate systems 
    into the `round_up_crs_list` as it will decrease precision.

    The function will roundup to the nearest meter if the GeoDataFrame `crs` 
    is in the `round_up_crs_list`.

    Args:
        gdf (gpd.GeoDataFrame): [description]
        round_up_crs_list (list, optional): [description]. Defaults to ['EPSG:3006', 'EPSG:3035'].

    Returns:
        List[Tuple[Point]]: [description]
    """

    assert 'geometry' in gdf.columns
    assert gdf.crs is not None
        
    # Try to assess if we need to round up the
    if gdf.crs in round_up_crs_list:
        # NOTE: we round-up to nearest meter
        # Check type of centroids, is it a GeoSeries of Points or just a List of Points
        centroids = gdf['geometry'].apply( lambda p: Point(int(round(p.centroid.x)), int(round(p.centroid.y))))        
    else:         
        centroids = gdf['geometry'].apply( lambda p: Point(p.centroid.x, p.centroid.y))
       
    return centroids


def load_single_raster(raster_filepath:str, *args, **kwargs):
    """utility generator function to load rasters using rasterio. 
    To be used in combination with decorators.

    Args:
        raster_filepath (str): [description]

    Yields:
        [Rasterio dataset]: Dataset
    """
    assert os.path.isfile(raster_filepath)

    with rasterio.open(fp=raster_filepath, *args, **kwargs) as dataset:
        yield dataset


def extract_raster_values(xy_tuples: List[Point]):
    """Decorator to extract one or multiple raster values based on a list of `Point` `xy_tuples`.
    Expects as dataset generated from the `load_single_raster` function.

    Args:
        xy_tuples (List[Point]): [description]

    Returns:
        [List]: List of values at xy_tuples sampled Points

    The Coordinate Reference System (CRS) of the `xy_tuple` must match rasters CRS.
    """

    assert xy_tuples is not None
    assert isinstance(xy_tuples, List[Point])
    # assert isinstance(xy_tuples[0], Point)

    def get_dataset_values(func):
        def wrapper(raster_filepath, *args, **kwargs):
            extracted_values = []
            for dataset in func(raster_filepath, *args, **kwargs):
                value_generator = dataset.sample(xy_tuples)
                try:
                    for values in value_generator:
                        _value = list(values)[0]
                        extracted_values.append(_value)
                        logging.info(values)
                except Exception as msg:
                    _value = None
                    logging.exception(msg)
                return extracted_values # _value
        return wrapper
    return get_dataset_values



def extract_raster_value_at_point(raster_filepath:str, xy_tuple:tuple)->Union[float,int]:
    """Extracts raster value at point with xy_tuple coordinates, using rasterio. 

    The Coordinate Reference System (CRS) of the `xy_tuple` must match rasters CRS.
    """
    assert os.path.isfile(raster_filepath)

    with rasterio.open(raster_filepath) as dataset:
        value_generator = dataset.sample([xy_tuple])

        try:
            for values in value_generator:
                extracted_value = list(values)[0]
        except Exception as msg:
            extracted_value = None
            logging.exception(msg)
        return extracted_value